/*
 * Copyright 2018-present MongoDB, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef KMS_MESSAGE_H
#define KMS_MESSAGE_H

#include <sys/types.h>

#include "kms_message_defines.h"
#include "kms_kmip_request.h"
#include "kms_kmip_response.h"
#include "kms_kmip_response_parser.h"
#include "kms_request_opt.h"
#include "kms_request.h"
#include "kms_response.h"
#include "kms_response_parser.h"
#include "kms_caller_identity_request.h"
#include "kms_decrypt_request.h"
#include "kms_encrypt_request.h"

#endif /* KMS_MESSAGE_H */
