/*
 * Copyright 2018-present MongoDB, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef KMS_REQUEST_OPT_PRIVATE_H
#define KMS_REQUEST_OPT_PRIVATE_H

#include "kms_message/kms_message_defines.h"
#include "kms_message/kms_request_opt.h"
#include "kms_crypto.h"

#include <stdbool.h>

struct _kms_request_opt_t {
   bool connection_close;
   _kms_crypto_t crypto;
   kms_request_provider_t provider;
};

#endif /* KMS_REQUEST_OPT_PRIVATE_H */
