/*
 * Copyright 2009-present MongoDB, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef TEST_MONGOC_RETRYABILITY_HELPERS_H
#define TEST_MONGOC_RETRYABILITY_HELPERS_H

#include <mongoc/mongoc-array-private.h>

mongoc_array_t
_test_get_mongos_clients (const char **ports, size_t num_ports);

#endif
