/*
 * Copyright 2009-present MongoDB, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef UNIFIED_OPERATION_H
#define UNIFIED_OPERATION_H

#include "runner.h"
#include "test-diagnostics.h"

#include <signal.h>

bool
operation_run (test_t *test, bson_t *bson, bson_error_t *error);

// Should only be used by test-atlas-executor.
extern volatile sig_atomic_t operation_loop_terminated;

#endif /* UNIFIED_OPERATION_H */
